@@ -1,4 +1,4 @@
-/* Copyright 2016-2019 Dimitrij Mijoski, Sander van Geloven
+/* Copyright 2016-2020 Dimitrij Mijoski, Sander van Geloven
  *
  * This file is part of Nuspell.
  *
@@ -18,25 +18,34 @@
 
 #include "dictionary.hxx"
 #include "finder.hxx"
-#include "utils.hxx"
 
+#include <cassert>
 #include <fstream>
 #include <iomanip>
 #include <iostream>
-
-#include <boost/locale.hpp>
-
-// manually define if not supplied by the build system
-#ifndef PROJECT_VERSION
-#define PROJECT_VERSION "unknown.version"
-#endif
-#define PACKAGE_STRING "nuspell " PROJECT_VERSION
+#include <unicode/brkiter.h>
+#include <unicode/ucnv.h>
 
 #if defined(__MINGW32__) || defined(__unix__) || defined(__unix) ||            \
     (defined(__APPLE__) && defined(__MACH__))
 #include <getopt.h>
 #include <unistd.h>
 #endif
+#ifdef _POSIX_VERSION
+#include <langinfo.h>
+#endif
+#ifdef _WIN32
+#include <io.h>
+#define NOMINMAX
+#define WIN32_LEAN_AND_MEAN
+#include <windows.h>
+#endif
+
+// manually define if not supplied by the build system
+#ifndef PROJECT_VERSION
+#define PROJECT_VERSION "unknown.version"
+#endif
+#define PACKAGE_STRING "nuspell " PROJECT_VERSION
 
 using namespace std;
 using namespace nuspell;
@@ -60,7 +69,7 @@ enum Mode {
 
 struct Args_t {
 	Mode mode = DEFAULT_MODE;
-	bool unicode_segmentation = false;
+	bool whitespace_segmentation = false;
 	string program_name = "nuspell";
 	string dictionary;
 	string encoding;
@@ -88,7 +97,7 @@ auto Args_t::parse_args(int argc, char* argv[]) -> void
 	int c;
 	// The program can run in various modes depending on the
 	// command line options. mode is FSM state, this while loop is FSM.
-	const char* shortopts = ":d:i:aDGLSlhv";
+	const char* shortopts = ":d:i:aDGLslhv";
 	const struct option longopts[] = {
 	    {"version", 0, nullptr, 'v'},
 	    {"help", 0, nullptr, 'h'},
@@ -152,8 +161,8 @@ auto Args_t::parse_args(int argc, char* argv[]) -> void
 				mode = ERROR_MODE;
 
 			break;
-		case 'S':
-			unicode_segmentation = true;
+		case 's':
+			whitespace_segmentation = true;
 
 			break;
 		case 'h':
@@ -193,67 +202,6 @@ auto Args_t::parse_args(int argc, char* argv[]) -> void
 #endif
 }
 
-class My_Dictionary : public Dictionary {
-	Hash_Multiset<string> personal;
-
-      public:
-	auto& operator=(const Dictionary& d)
-	{
-		static_cast<Dictionary&>(*this) = d;
-		return *this;
-	}
-	auto& operator=(Dictionary&& d)
-	{
-		static_cast<Dictionary&>(*this) = move(d);
-		return *this;
-	}
-	auto spell(const string& word) const
-	{
-		auto correct = Dictionary::spell(word);
-		if (correct)
-			return true;
-		auto r = personal.equal_range(word);
-		correct = r.first != r.second;
-		return correct;
-	}
-	auto parse_personal_dict(istream& in, const locale& external_locale)
-	{
-		auto word = string();
-		auto wide_word = wstring();
-		while (getline(in, word)) {
-			auto ok = utf8_to_wide(word, wide_word);
-			ok &= to_narrow(wide_word, word, external_locale);
-			if (!ok)
-				continue;
-			personal.insert(word);
-		}
-		return in.eof();
-	}
-	auto parse_personal_dict(std::string name,
-	                         const locale& external_locale)
-	{
-#ifdef _WIN32
-		auto const PATH_SEPS = "\\/";
-#else
-		auto const PATH_SEPS = '/';
-#endif
-		auto file = ifstream();
-		auto path_sep_idx = name.find_last_of(PATH_SEPS);
-		if (path_sep_idx != name.npos)
-			name.erase(0, path_sep_idx + 1);
-		name.insert(0, ".nuspell_");
-		auto home = getenv("HOME");
-		if (home) {
-			name.insert(0, "/");
-			name.insert(0, home);
-		}
-		file.open(name);
-		if (file.is_open())
-			return parse_personal_dict(file, external_locale);
-		return true;
-	}
-};
-
 /**
  * @brief Prints help information to standard output.
  *
@@ -265,8 +213,8 @@ auto print_help(const string& program_name) -> void
 	auto& o = cout;
 	o << "Usage:\n"
 	     "\n";
-	o << p << " [-S] [-d dict_NAME] [-i enc] [file_name]...\n";
-	o << p << " -l|-G [-L] [-S] [-d dict_NAME] [-i enc] [file_name]...\n";
+	o << p << " [-s] [-d dict_NAME] [-i enc] [file_name]...\n";
+	o << p << " -l|-G [-L] [-s] [-d dict_NAME] [-i enc] [file_name]...\n";
 	o << p << " -D|-h|--help|-v|--version\n";
 	o << "\n"
 	     "Check spelling of each FILE. Without FILE, check standard "
@@ -280,7 +228,10 @@ auto print_help(const string& program_name) -> void
 	     "  -l            print only misspelled words or lines\n"
 	     "  -G            print only correct words or lines\n"
 	     "  -L            lines mode\n"
-	     "  -S            use Unicode text segmentation to extract words\n"
+	     "  -s            use simple whitespace text segmentation to\n"
+	     "                extract words instead of the default Unicode\n"
+	     "                text segmentation. It is not recommended to use\n"
+	     "                this.\n"
 	     "  -h, --help    print this help and exit\n"
 	     "  -v, --version print version number and exit\n"
 	     "\n";
@@ -299,15 +250,14 @@ auto print_version() -> void
 {
 	cout << PACKAGE_STRING
 	    "\n"
-	    "Copyright (C) 2017-2018 Dimitrij Mijoski and Sander van Geloven\n"
+	    "Copyright (C) 2016-2020 Dimitrij Mijoski and Sander van Geloven\n"
 	    "License LGPLv3+: GNU LGPL version 3 or later "
 	    "<http://gnu.org/licenses/lgpl.html>.\n"
 	    "This is free software: you are free to change and "
 	    "redistribute it.\n"
 	    "There is NO WARRANTY, to the extent permitted by law.\n"
 	    "\n"
-	    "Written by Dimitrij Mijoski, Sander van Geloven and others,\n"
-	    "see https://github.com/nuspell/nuspell/blob/master/AUTHORS\n";
+	    "Written by Dimitrij Mijoski and Sander van Geloven.\n";
 }
 
 /**
@@ -315,7 +265,7 @@ auto print_version() -> void
  *
  * @param f a finder for search paths and located dictionary.
  */
-auto list_dictionaries(const Finder& f) -> void
+auto list_dictionaries(const Dict_Finder_For_CLI_Tool& f) -> void
 {
 	if (f.get_dir_paths().empty()) {
 		cout << "No search paths available" << '\n';
@@ -341,14 +291,57 @@ auto list_dictionaries(const Finder& f) -> void
 	}
 }
 
-auto process_word(
-    Mode mode, const My_Dictionary& dic, const string& line, streampos pos_line,
-    string::const_iterator b, string::const_iterator c, bool tellg_supported,
-    string& word,
-    vector<pair<string::const_iterator, string::const_iterator>>& wrong_words,
-    vector<string>& suggestions, ostream& out)
+auto to_utf8(string_view source, string& dest, UConverter* ucnv,
+             UErrorCode& uerr)
+{
+	dest.resize(dest.capacity());
+	auto len = ucnv_toAlgorithmic(UCNV_UTF8, ucnv, dest.data(), dest.size(),
+	                              source.data(), source.size(), &uerr);
+	dest.resize(len);
+	if (uerr == U_BUFFER_OVERFLOW_ERROR) {
+		uerr = U_ZERO_ERROR;
+		ucnv_toAlgorithmic(UCNV_UTF8, ucnv, dest.data(), dest.size(),
+		                   source.data(), source.size(), &uerr);
+	}
+}
+
+auto from_utf8(string_view source, string& dest, UConverter* ucnv,
+               UErrorCode& uerr)
+{
+	dest.resize(dest.capacity());
+	auto len =
+	    ucnv_fromAlgorithmic(ucnv, UCNV_UTF8, dest.data(), dest.size(),
+	                         source.data(), source.size(), &uerr);
+	dest.resize(len);
+	if (uerr == U_BUFFER_OVERFLOW_ERROR) {
+		uerr = U_ZERO_ERROR;
+		ucnv_fromAlgorithmic(ucnv, UCNV_UTF8, dest.data(), dest.size(),
+		                     source.data(), source.size(), &uerr);
+	}
+}
+
+auto to_unicode_string(string_view source, icu::UnicodeString& dest,
+                       UConverter* ucnv, UErrorCode& uerr)
+{
+	auto buf = dest.getBuffer(-1);
+	auto len = ucnv_toUChars(ucnv, buf, dest.getCapacity(), source.data(),
+	                         source.size(), &uerr);
+	if (uerr == U_BUFFER_OVERFLOW_ERROR) {
+		uerr = U_ZERO_ERROR;
+		dest.releaseBuffer(0);
+		buf = dest.getBuffer(len);
+		if (!buf)
+			throw bad_alloc();
+		len = ucnv_toUChars(ucnv, buf, dest.getCapacity(),
+		                    source.data(), source.size(), &uerr);
+	}
+	dest.releaseBuffer(len);
+}
+
+auto process_word(Mode mode, const Dictionary& dic, string_view word,
+                  size_t pos_word, vector<string_view>& wrong_words,
+                  vector<string>& suggestions, ostream& out)
 {
-	word.assign(b, c);
 	auto correct = dic.spell(word);
 	switch (mode) {
 	case DEFAULT_MODE: {
@@ -357,9 +350,6 @@ auto process_word(
 			break;
 		}
 		dic.suggest(word, suggestions);
-		auto pos_word = pos_line;
-		if (tellg_supported)
-			pos_word += b - begin(line);
 		if (suggestions.empty()) {
 			out << "# " << word << ' ' << pos_word << '\n';
 			break;
@@ -383,20 +373,71 @@ auto process_word(
 	case MISSPELLED_LINES_MODE:
 	case CORRECT_LINES_MODE:
 		if (!correct)
-			wrong_words.emplace_back(b, c);
+			wrong_words.push_back(word);
 		break;
 	default:
 		break;
 	}
 }
 
-auto process_line(
-    Mode mode, const string& line,
-    const vector<pair<string::const_iterator, string::const_iterator>>&
-        wrong_words,
-    ostream& out)
+auto process_word_other_encoding(Mode mode, const Dictionary& dic,
+                                 string_view word, string_view u8word,
+                                 size_t pos_word,
+                                 vector<string_view>& wrong_words,
+                                 vector<string>& suggestions, ostream& out,
+                                 UConverter* ucnv, UErrorCode& uerr)
+{
+	auto correct = dic.spell(u8word);
+	switch (mode) {
+	case DEFAULT_MODE: {
+		if (correct) {
+			out << "*\n";
+			break;
+		}
+		dic.suggest(u8word, suggestions);
+		if (suggestions.empty()) {
+			out << "# " << word << ' ' << pos_word << '\n';
+			break;
+		}
+		out << "& " << word << ' ' << suggestions.size() << ' '
+		    << pos_word << ": ";
+		auto sug_in_encoding = string();
+		from_utf8(suggestions[0], sug_in_encoding, ucnv, uerr);
+		out << sug_in_encoding;
+		for_each(begin(suggestions) + 1, end(suggestions),
+		         [&](const string& u8sug) {
+			         out << ", ";
+			         from_utf8(u8sug, sug_in_encoding, ucnv, uerr);
+			         out << sug_in_encoding;
+		         });
+		out << '\n';
+		break;
+	}
+	case MISSPELLED_WORDS_MODE:
+		if (!correct)
+			out << word << '\n';
+		break;
+	case CORRECT_WORDS_MODE:
+		if (correct)
+			out << word << '\n';
+		break;
+	case MISSPELLED_LINES_MODE:
+	case CORRECT_LINES_MODE:
+		if (!correct)
+			wrong_words.push_back(word);
+		break;
+	default:
+		break;
+	}
+}
+
+auto finish_line(Mode mode, const string& line,
+                 const vector<string_view>& wrong_words, ostream& out)
 {
 	switch (mode) {
+	case DEFAULT_MODE:
+		out << '\n';
+		break;
 	case MISSPELLED_LINES_MODE:
 		if (!wrong_words.empty())
 			out << line << '\n';
@@ -411,132 +452,147 @@ auto process_line(
 }
 
 auto whitespace_segmentation_loop(istream& in, ostream& out,
-                                  const My_Dictionary& dic, Mode mode)
+                                  const Dictionary& dic, Mode mode,
+                                  UConverter* ucnv, UErrorCode& uerr)
 {
 	auto line = string();
-	auto word = string();
 	auto suggestions = vector<string>();
-	using Str_Iter = string::const_iterator;
-	auto wrong_words = vector<pair<Str_Iter, Str_Iter>>();
+	auto wrong_words = vector<string_view>();
 	auto loc = in.getloc();
-	auto pos_line = in.tellg();
-	auto tellg_supported = true;
-	if (pos_line < 0) {
-		pos_line = 0;
-		tellg_supported = false;
-	}
-	auto line_num = size_t(0);
 	auto& facet = use_facet<ctype<char>>(loc);
 	auto isspace = [&](char c) { return facet.is(facet.space, c); };
+	auto u8word = string();
+	auto is_utf8 = ucnv_getType(ucnv) == UCNV_UTF8;
+
 	while (getline(in, line)) {
-		++line_num;
 		wrong_words.clear();
 		for (auto a = begin(line); a != end(line);) {
-			auto b = find_if_not(a, end(line), isspace);
-			if (b == end(line))
+			a = find_if_not(a, end(line), isspace);
+			if (a == end(line))
 				break;
-			auto c = find_if(b, end(line), isspace);
-
-			process_word(mode, dic, line, pos_line, b, c,
-			             tellg_supported, word, wrong_words,
-			             suggestions, out);
-
-			a = c;
+			auto b = find_if(a, end(line), isspace);
+			auto word = string_view(&*a, distance(a, b));
+			auto pos_word = distance(begin(line), a);
+			if (is_utf8) {
+				process_word(mode, dic, word, pos_word,
+				             wrong_words, suggestions, out);
+			}
+			else {
+				to_utf8(word, u8word, ucnv, uerr);
+				process_word_other_encoding(
+				    mode, dic, word, u8word, pos_word,
+				    wrong_words, suggestions, out, ucnv, uerr);
+			}
+			a = b;
 		}
-		process_line(mode, line, wrong_words, out);
-
-		if (tellg_supported)
-			pos_line = in.tellg();
+		finish_line(mode, line, wrong_words, out);
 	}
 }
 
-auto unicode_segentation_loop(istream& in, ostream& out,
-                              const My_Dictionary& dic, Mode mode)
+auto is_word_break(int32_t typ)
+{
+	return (UBRK_WORD_NUMBER <= typ && typ < UBRK_WORD_NUMBER_LIMIT) ||
+	       (UBRK_WORD_LETTER <= typ && typ < UBRK_WORD_LETTER_LIMIT) ||
+	       (UBRK_WORD_KANA <= typ && typ < UBRK_WORD_KANA_LIMIT) ||
+	       (UBRK_WORD_IDEO <= typ && typ < UBRK_WORD_IDEO_LIMIT);
+}
+
+auto segment_line_utf8(Mode mode, const Dictionary& dic, const string& line,
+                       UText* utext, icu::BreakIterator* ubrkiter,
+                       UErrorCode& uerr, vector<string>& suggestions,
+                       vector<string_view>& wrong_words, ostream& out)
+{
+	utext_openUTF8(utext, line.data(), line.size(), &uerr);
+	ubrkiter->setText(utext, uerr);
+	for (auto i = ubrkiter->first(), prev = 0; i != ubrkiter->DONE;
+	     prev = i, i = ubrkiter->next()) {
+		auto typ = ubrkiter->getRuleStatus();
+		if (is_word_break(typ)) {
+			auto word = string_view(line).substr(prev, i - prev);
+			process_word(mode, dic, word, prev, wrong_words,
+			             suggestions, out);
+		}
+	}
+	finish_line(mode, line, wrong_words, out);
+	assert(U_SUCCESS(uerr));
+}
+
+auto segment_line_generic(Mode mode, const Dictionary& dic, const string& line,
+                          icu::UnicodeString& uline, UConverter* ucnv,
+                          icu::BreakIterator* ubrkiter, UErrorCode& uerr,
+                          string& u8word, vector<string>& suggestions,
+                          vector<string_view>& wrong_words, ostream& out)
+{
+	to_unicode_string(line, uline, ucnv, uerr);
+	ubrkiter->setText(uline);
+	size_t orig_prev = 0, orig_i = 0;
+	auto src = line.c_str();
+	auto src_end = src + line.size();
+
+	ucnv_resetToUnicode(ucnv);
+	for (auto i = ubrkiter->first(), prev = 0; i != ubrkiter->DONE;
+	     prev = i, i = ubrkiter->next(), orig_prev = orig_i) {
+
+		for (auto j = prev; j != i; ++j) {
+			auto cp = ucnv_getNextUChar(ucnv, &src, src_end, &uerr);
+
+			// U_IS_SURROGATE(uline[j]) or
+			// U_IS_LEAD(uline[j]) can work too
+			j += !U_IS_BMP(cp);
+		}
+		orig_i = distance(line.c_str(), src);
+
+		auto typ = ubrkiter->getRuleStatus();
+		if (is_word_break(typ)) {
+			auto uword = uline.tempSubStringBetween(prev, i);
+			u8word.clear();
+			uword.toUTF8String(u8word);
+			auto word = string_view(line).substr(
+			    orig_prev, orig_i - orig_prev);
+			process_word_other_encoding(
+			    mode, dic, word, u8word, orig_prev, wrong_words,
+			    suggestions, out, ucnv, uerr);
+		}
+	}
+	finish_line(mode, line, wrong_words, out);
+	assert(U_SUCCESS(uerr));
+}
+
+auto unicode_segentation_loop(istream& in, ostream& out, const Dictionary& dic,
+                              Mode mode, UConverter* ucnv, UErrorCode& uerr)
 {
-	namespace b = boost::locale::boundary;
 	auto line = string();
-	auto word = string();
 	auto suggestions = vector<string>();
-	using Str_Iter = string::const_iterator;
-	auto wrong_words = vector<pair<Str_Iter, Str_Iter>>();
-	auto loc = in.getloc();
-	auto pos_line = in.tellg();
-	auto tellg_supported = true;
-	if (pos_line < 0) {
-		pos_line = 0;
-		tellg_supported = false;
-	}
-	auto line_num = size_t(0);
-	auto index = b::ssegment_index();
-	index.rule(b::word_any);
-	auto line_stream = istringstream();
+	auto wrong_words = vector<string_view>();
+
+	// TODO: try to use Locale constructed from dictionary name.
+	auto ubrkiter = unique_ptr<icu::BreakIterator>(
+	    icu::BreakIterator::createWordInstance(icu::Locale(), uerr));
+	auto utext = icu::LocalUTextPointer(
+	    utext_openUTF8(nullptr, line.data(), line.size(), &uerr));
+	auto uline = icu::UnicodeString();
+	auto u8word = string();
+	auto is_utf8 = ucnv_getType(ucnv) == UCNV_UTF8;
+
 	while (getline(in, line)) {
-		++line_num;
-		index.map(b::word, begin(line), end(line), loc);
 		wrong_words.clear();
-		auto a = cbegin(line);
-		for (auto& segment : index) {
-			auto b = begin(segment);
-			auto c = end(segment);
-
-			process_word(mode, dic, line, pos_line, b, c,
-			             tellg_supported, word, wrong_words,
-			             suggestions, out);
-
-			a = c;
-		}
-		process_line(mode, line, wrong_words, out);
-
-		if (tellg_supported)
-			pos_line = in.tellg();
+		if (is_utf8)
+			segment_line_utf8(mode, dic, line, utext.getAlias(),
+			                  ubrkiter.get(), uerr, suggestions,
+			                  wrong_words, out);
+		else
+			segment_line_generic(mode, dic, line, uline, ucnv,
+			                     ubrkiter.get(), uerr, u8word,
+			                     suggestions, wrong_words, out);
 	}
 }
 
-namespace std {
-ostream& operator<<(ostream& out, const locale& loc)
-{
-	if (has_facet<boost::locale::info>(loc)) {
-		auto& f = use_facet<boost::locale::info>(loc);
-		out << "name=" << f.name() << ", lang=" << f.language()
-		    << ", country=" << f.country() << ", enc=" << f.encoding();
-	}
-	else {
-		out << loc.name();
-	}
-	return out;
-}
-} // namespace std
-
 int main(int argc, char* argv[])
 {
 	// May speed up I/O. After this, don't use C printf, scanf etc.
 	ios_base::sync_with_stdio(false);
 
 	auto args = Args_t(argc, argv);
-	if (args.mode == ERROR_MODE) {
-		cerr << "Invalid (combination of) arguments, try '"
-		     << args.program_name << " --help' for more information\n";
-		return 1;
-	}
-	boost::locale::generator gen;
-	auto loc = std::locale();
-	try {
-		if (args.encoding.empty())
-			loc = gen("");
-		else
-			loc = gen("en_US." + args.encoding);
-	}
-	catch (const boost::locale::conv::invalid_charset_error& e) {
-		cerr << e.what() << '\n';
-#ifdef _POSIX_VERSION
-		cerr << "Nuspell error: see `locale -m` for supported "
-		        "encodings.\n";
-#endif
-		return 1;
-	}
-	cin.imbue(loc);
-	cout.imbue(loc);
-
 	switch (args.mode) {
 	case HELP_MODE:
 		print_help(args.program_name);
@@ -544,26 +600,64 @@ int main(int argc, char* argv[])
 	case VERSION_MODE:
 		print_version();
 		return 0;
+	case ERROR_MODE:
+		cerr << "Invalid (combination of) arguments, try '"
+		     << args.program_name << " --help' for more information\n";
+		return 1;
 	default:
 		break;
 	}
-	clog << "INFO: I/O  locale " << loc << '\n';
-
-	auto f = Finder::search_all_dirs_for_dicts();
-
+	auto f = Dict_Finder_For_CLI_Tool();
 	if (args.mode == LIST_DICTIONARIES_MODE) {
 		list_dictionaries(f);
 		return 0;
 	}
+	char* loc_str = nullptr;
+#ifdef _WIN32
+	loc_str = setlocale(LC_CTYPE, nullptr); // will return "C"
+
+	/* On Windows, the console is a buggy thing. If the default C locale is
+	active, then the encoding of the strings gotten from C or C++ stdio
+	(fgets, scanf, cin) is GetConsoleCP(). Stdout accessed via standard
+	functions (printf, cout) expects encoding of GetConsoleOutputCP() which
+	is the same as GetConsoleCP() unless manually changed. By default both
+	are the active OEM encoding, unless changed with the command chcp, or by
+	calling the Set functions.
+
+	If we call setlocale(LC_CTYPE, ""), or let's say setlocale(LC_CTYPE,
+	".1251"), then stdin will still return in the encoding GetConsoleCP(),
+	but stdout functions like printf now will expect a different encoding,
+	the one set via setlocale. Because of this mess don't change locale with
+	setlocale on Windows.
+
+	When stdin or stout are redirected from/to file or another terminal like
+	the one in MSYS2, they are read/written as-is. Then we will assume UTF-8
+	encoding. */
+#else
+	loc_str = setlocale(LC_CTYPE, "");
+	if (!loc_str) {
+		clog << "WARNING: Invalid locale string, fall back to \"C\".\n";
+		loc_str = setlocale(LC_CTYPE, nullptr); // will return "C"
+	}
+#endif
+	auto loc_str_sv = string_view(loc_str);
+	if (args.encoding.empty()) {
+#if _POSIX_VERSION
+		auto enc_str = nl_langinfo(CODESET);
+		args.encoding = enc_str;
+#elif _WIN32
+		if (_isatty(_fileno(stdin)) || _isatty(_fileno(stdout)))
+			args.encoding = "cp" + to_string(GetConsoleCP());
+		else
+			args.encoding = "UTF-8";
+#endif
+	}
+	clog << "INFO: Locale LC_CTYPE=" << loc_str_sv
+	     << ", Used encoding=" << args.encoding << '\n';
 	if (args.dictionary.empty()) {
 		// infer dictionary from locale
-		auto& info = use_facet<boost::locale::info>(loc);
-		args.dictionary = info.language();
-		auto c = info.country();
-		if (!c.empty()) {
-			args.dictionary += '_';
-			args.dictionary += c;
-		}
+		auto idx = min(loc_str_sv.find('.'), loc_str_sv.find('@'));
+		args.dictionary = loc_str_sv.substr(0, idx);
 	}
 	if (args.dictionary.empty()) {
 		cerr << "No dictionary provided and can not infer from OS "
@@ -575,21 +669,39 @@ int main(int argc, char* argv[])
 		return 1;
 	}
 	clog << "INFO: Pointed dictionary " << filename << ".{dic,aff}\n";
-	auto dic = My_Dictionary();
+	auto dic = Dictionary();
 	try {
 		dic = Dictionary::load_from_path(filename);
-		dic.parse_personal_dict(args.dictionary, loc);
 	}
 	catch (const Dictionary_Loading_Error& e) {
 		cerr << e.what() << '\n';
 		return 1;
 	}
-	dic.imbue(loc);
-	auto loop_function = whitespace_segmentation_loop;
-	if (args.unicode_segmentation)
-		loop_function = unicode_segentation_loop;
+	// ICU reports all types of errors, logic errors and runtime errors
+	// using this enum. We should not check for logic errors, they should
+	// not happend. Optionally, only assert that they are not there can be
+	// used. We should check for runtime errors.
+	// The encoding conversion is a common case where runtime error can
+	// happen, but by default ICU uses Unicode replacement character on
+	// errors and reprots success. This can be changed, but there is no need
+	// for that.
+	auto uerr = U_ZERO_ERROR;
+	auto enc_cstr = args.encoding.c_str();
+	if (args.encoding.empty()) {
+		enc_cstr = nullptr;
+		clog << "WARNING: using default ICU encoding converter for IO"
+		     << endl;
+	}
+	auto ucnv = icu::LocalUConverterPointer(ucnv_open(enc_cstr, &uerr));
+	if (U_FAILURE(uerr)) {
+		cerr << "ERROR: Invalid encoding " << args.encoding << ".\n";
+		return 1;
+	}
+	auto loop_function = unicode_segentation_loop;
+	if (args.whitespace_segmentation)
+		loop_function = whitespace_segmentation_loop;
 	if (args.files.empty()) {
-		loop_function(cin, cout, dic, args.mode);
+		loop_function(cin, cout, dic, args.mode, ucnv.getAlias(), uerr);
 	}
 	else {
 		for (auto& file_name : args.files) {
@@ -598,8 +710,8 @@ int main(int argc, char* argv[])
 				cerr << "Can't open " << file_name << '\n';
 				return 1;
 			}
-			in.imbue(loc);
-			loop_function(in, cout, dic, args.mode);
+			loop_function(in, cout, dic, args.mode, ucnv.getAlias(),
+			              uerr);
 		}
 	}
 	return 0;
